import dataclasses
import json
import logging
import typing as t

import ase
import numpy as np
import socketio.exceptions
import tqdm
import znjson
import znsocket
from redis import Redis

from zndraw.base import Extension, ZnDrawBase
from zndraw.bonds import ASEComputeBonds
from zndraw.config import ArrowsConfig, ZnDrawConfig
from zndraw.draw import Geometry, Object3D
from zndraw.type_defs import (
    ATOMS_LIKE,
    CameraData,
    JupyterConfig,
    RegisterModifier,
    TimeoutConfig,
)
from zndraw.utils import ASEConverter, call_with_retry, emit_with_retry

log = logging.getLogger(__name__)


def _register_modifier(vis: "ZnDraw", data: RegisterModifier) -> None:
    log.debug(f"Registering modifier `{data['cls'].__name__}`")
    vis.socket.emit(
        "modifier:register",
        {
            "schema": data["cls"].model_json_schema(),
            "name": data["cls"].__name__,
            "public": data["public"],
            "timeout": data["timeout"],
        },
    )
    if vis._available:
        vis.socket.emit("modifier:available", list(vis._modifiers))


@dataclasses.dataclass
class ZnDraw(ZnDrawBase):
    url: str
    token: str | None = None
    auth_token: str | None = None

    socket: socketio.Client = dataclasses.field(
        default_factory=socketio.Client, repr=False
    )
    timeout: TimeoutConfig = dataclasses.field(
        default_factory=lambda: TimeoutConfig(
            connection=10,
            modifier=0.25,
            between_calls=0.1,
            emit_retries=3,
            call_retries=3,
            connect_retries=3,
        )
    )
    jupyter_config: JupyterConfig = dataclasses.field(
        default_factory=lambda: JupyterConfig(
            width="70%",
            height=600,
        )
    )

    maximum_message_size: int = dataclasses.field(default=500_000, repr=False)

    _modifiers: dict[str, RegisterModifier] = dataclasses.field(default_factory=dict)
    _available: bool = True

    bond_calculator: ASEComputeBonds | None = dataclasses.field(
        default_factory=ASEComputeBonds, repr=False
    )

    def __post_init__(self):
        def on_wakeup():
            if self._available:
                self.socket.emit("modifier:available", list(self._modifiers))

        self.url = self.url.replace("http", "ws")
        self.socket.on("connect", self._on_connect)
        self.socket.on("modifier:run", self._run_modifier)
        self.socket.on("modifier:wakeup", on_wakeup)
        self.socket.on("room:log", lambda x: print(x))

        for idx in range(self.timeout["connect_retries"] + 1):
            try:
                self.socket.connect(self.url, wait_timeout=self.timeout["connection"])
                break
            except socketio.exceptions.ConnectionError as err:
                log.warning("Connection failed. Retrying...")
                self.socket.sleep(self.timeout["connection"])
                if idx == self.timeout["connect_retries"]:
                    raise socketio.exceptions.ConnectionError(
                        f"Unable to connect to ZnDraw server at '{self.url}'. Is the server running?"
                    ) from err

    def _on_connect(self):
        log.debug("Connected to ZnDraw server")

        emit_with_retry(
            self.socket,
            "join",
            {
                "token": str(self.token),
                "auth_token": self.auth_token,
            },
            retries=self.timeout["emit_retries"],
        )

        for data in self._modifiers.values():
            _register_modifier(self, data)

    def __getitem__(self, index) -> ase.Atoms | list[ase.Atoms]:
        single_item = isinstance(index, int)
        if single_item:
            index = [index]

        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))

        # make negative indices count from the end
        index = [i if i >= 0 else len(self) + i for i in index]

        if any(x >= len(self) for x in index):
            raise IndexError("Index out of range")

        data = call_with_retry(
            self.socket,
            "room:frames:get",
            index,
            retries=self.timeout["call_retries"],
        )

        structures = [
            znjson.loads(
                json.dumps(x), cls=znjson.ZnDecoder.from_converters([ASEConverter])
            )
            for x in data.values()
        ]
        return structures[0] if single_item else structures

    def __setitem__(
        self, index: int | list[int] | slice, value: ATOMS_LIKE | list[ATOMS_LIKE]
    ):
        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))
        if isinstance(index, int):
            index = [index]
            value = [value]

        data = {}
        for i, val in zip(index, value):
            if isinstance(val, ase.Atoms):
                if not hasattr(val, "connectivity") and self.bond_calculator is not None:
                    val.connectivity = self.bond_calculator.get_bonds(val)
                data[i] = znjson.dumps(
                    val, cls=znjson.ZnEncoder.from_converters([ASEConverter])
                )
            else:
                data[i] = val
            if '"_type": "ase.Atoms"' not in data[i]:
                raise ValueError("Unable to parse provided data object")
        call_with_retry(
            self.socket,
            "room:frames:set",
            data,
            retries=self.timeout["call_retries"],
        )

    def __len__(self) -> int:
        return call_with_retry(
            self.socket,
            "room:length:get",
            retries=self.timeout["call_retries"],
        )

    def __delitem__(self, index: int | slice | list[int]):
        if isinstance(index, int):
            index = [index]
        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))

        call_with_retry(
            self.socket,
            "room:frames:delete",
            index,
            retries=self.timeout["call_retries"],
        )

    def _repr_html_(self):
        from IPython.display import IFrame

        address = f"{self.url}/token/{self.token}"
        # TODO: save address and do not replace in post_init
        address = address.replace("ws", "http")
        log.info(f"Opening ZnDraw at {address}")
        return IFrame(
            address,
            width=self.jupyter_config["width"],
            height=self.jupyter_config["height"],
        )._repr_html_()

    def insert(self, index: int, value: ATOMS_LIKE):
        if isinstance(value, ase.Atoms):
            if not hasattr(value, "connectivity") and self.bond_calculator is not None:
                value.connectivity = self.bond_calculator.get_bonds(value)

            value = znjson.dumps(
                value, cls=znjson.ZnEncoder.from_converters([ASEConverter])
            )

        if '"_type": "ase.Atoms"' not in value:
            raise ValueError("Unable to parse provided data object")

        call_with_retry(
            self.socket,
            "room:frames:insert",
            {
                "index": index,
                "value": value,
            },
            retries=self.timeout["call_retries"],
        )

    def extend(self, values: list[ATOMS_LIKE]):
        msg = {}

        # enable tbar if more than 10 messages are sent
        # approximated by the size of the first frame

        if not isinstance(values, list):
            raise ValueError("Unable to parse provided data object")

        show_tbar = (
            len(values)
            * len(
                znjson.dumps(
                    values[0], cls=znjson.ZnEncoder.from_converters([ASEConverter])
                ).encode("utf-8")
            )
        ) > (10 * self.maximum_message_size)
        tbar = tqdm.tqdm(
            values, desc="Sending frames", unit=" frame", disable=not show_tbar
        )

        for i, val in enumerate(tbar, start=len(self)):
            if isinstance(val, ase.Atoms):
                if not hasattr(val, "connectivity") and self.bond_calculator is not None:
                    val.connectivity = self.bond_calculator.get_bonds(val)

                msg[i] = znjson.dumps(
                    val, cls=znjson.ZnEncoder.from_converters([ASEConverter])
                )
            else:
                msg[i] = val
            if '"_type": "ase.Atoms"' not in msg[i]:
                raise ValueError("Unable to parse provided data object")
            if len(json.dumps(msg).encode("utf-8")) > self.maximum_message_size:
                call_with_retry(
                    self.socket,
                    "room:frames:set",
                    msg,
                    retries=self.timeout["call_retries"],
                )
                msg = {}
                # after each large message, wait a bit
                self.socket.sleep(self.timeout["modifier"])
        if len(msg) > 0:  # Only send the message if it's not empty
            call_with_retry(
                self.socket,
                "room:frames:set",
                msg,
                retries=self.timeout["call_retries"],
            )

    @property
    def selection(self) -> list[int]:
        return call_with_retry(
            self.socket,
            "room:selection:get",
            retries=self.timeout["call_retries"],
        )["0"]

    @selection.setter
    def selection(self, value: list[int]):
        if not isinstance(value, list):
            raise ValueError("Selection must be a list")
        if not all(isinstance(x, int) for x in value):
            raise ValueError("Selection must be a list of integers")
        if len(value) != len(set(value)):
            raise ValueError("Selection must not contain duplicates")

        max_index = len(self.atoms)
        if any(x >= max_index for x in value):
            raise IndexError("Selection out of range")
        if any(x < 0 for x in value):
            raise IndexError("Selection must be positive")
        emit_with_retry(
            self.socket,
            "room:selection:set",
            {"0": value},
            retries=self.timeout["emit_retries"],
        )

    @property
    def step(self) -> int:
        return int(
            call_with_retry(
                self.socket,
                "room:step:get",
                retries=self.timeout["call_retries"],
            )
        )

    def log(self, message: str) -> None:
        emit_with_retry(
            self.socket,
            "room:log",
            message,
            retries=self.timeout["emit_retries"],
        )

    @step.setter
    def step(self, value: int):
        if not isinstance(value, int):
            raise ValueError("Step must be an integer")
        if value < 0:
            raise ValueError("Step must be positive")
        if value >= len(self):
            raise IndexError("Step out of range")
        # Have only one step per room!
        # shared rooms are rare anyhow and making per-client steps
        # and room hosts is annoying
        # what about the camera?
        # or collect the steps of all clients in a dict
        # and save the host and go from there, also fine and not too much worker.
        emit_with_retry(
            self.socket,
            "room:step:set",
            value,
            retries=self.timeout["emit_retries"],
        )

    @property
    def figure(self) -> str:
        return call_with_retry(
            self.socket,
            "analysis:figure:get",
            retries=self.timeout["call_retries"],
        )

    @figure.setter
    def figure(self, fig: str) -> None:
        emit_with_retry(
            self.socket,
            "analysis:figure:set",
            fig,
            retries=self.timeout["emit_retries"],
        )

    @property
    def atoms(self) -> ase.Atoms:
        return self[self.step]

    @atoms.setter
    def atoms(self, atoms: ATOMS_LIKE) -> None:
        self[[self.step]] = [atoms]

    @property
    def points(self) -> np.ndarray:
        return np.array(
            call_with_retry(
                self.socket,
                "room:points:get",
                retries=self.timeout["call_retries"],
            )["0"]
        )

    @points.setter
    def points(self, points: np.ndarray | list) -> None:
        if isinstance(points, list):
            points = np.array(points)
        emit_with_retry(
            self.socket,
            "room:points:set",
            {"0": points.tolist()},
            retries=self.timeout["emit_retries"],
        )

    @property
    def bookmarks(self) -> dict[int, str]:
        return {
            int(k): v
            for k, v in call_with_retry(
                self.socket,
                "room:bookmarks:get",
                retries=self.timeout["call_retries"],
            ).items()
        }

    @bookmarks.setter
    def bookmarks(self, value: dict[int, str]):
        if not isinstance(value, dict):
            raise ValueError("Bookmarks must be a dictionary")
        if not all(isinstance(x, int) for x in value.keys()):
            raise ValueError("Bookmark keys must be integers")
        if not all(isinstance(x, str) for x in value.values()):
            raise ValueError("Bookmark values must be strings")

        emit_with_retry(
            self.socket,
            "room:bookmarks:set",
            value,
            retries=self.timeout["emit_retries"],
        )

    @property
    def camera(self) -> CameraData:
        return call_with_retry(
            self.socket,
            "room:camera:get",
            retries=self.timeout["call_retries"],
        )

    @camera.setter
    def camera(self, value: CameraData):
        if set(value.keys()) != {"position", "target"}:
            raise ValueError("Camera must have 'position' and 'target' keys")
        self.socket.emit("room:camera:set", {"content": value, "emit": True})

    @property
    def geometries(self) -> list[Object3D]:
        return [
            Geometry(method=x).method
            for x in call_with_retry(
                self.socket,
                "room:geometry:get",
                retries=self.timeout["call_retries"],
            )
        ]

    @geometries.setter
    def geometries(self, value: list[Object3D]):
        if not isinstance(value, list):
            raise ValueError("Geometries must be a list")
        if not all(isinstance(x, Object3D) for x in value):
            raise ValueError("Geometries must be a list of Object3D instances")
        emit_with_retry(
            self.socket,
            "room:geometry:set",
            [x.model_dump() for x in value],
            retries=self.timeout["emit_retries"],
        )

    @property
    def config(self) -> ZnDrawConfig:
        config: dict = call_with_retry(
            self.socket,
            "room:config:get",
            retries=self.timeout["call_retries"],
        )
        return ZnDrawConfig(vis=self, arrows=ArrowsConfig(**config["arrows"]))

    @property
    def locked(self) -> bool:
        return call_with_retry(self.socket, "room:lock:get")

    @locked.setter
    def locked(self, value: bool) -> None:
        emit_with_retry(self.socket, "room:lock:set", value)

    def register_modifier(
        self,
        cls: t.Type[Extension],
        run_kwargs: dict | None = None,
        public: bool = False,
        timeout: float = 60,
        use_frozen: bool = True,
    ):
        """Register a modifier class.

        Attributes
        ----------
        cls : UpdateScene
            The modifier class to register.
        run_kwargs : dict, optional
            Keyword arguments to pass to the run method of the modifier class.
        public : bool, optional
            Whether to enable the modifier for ALL sessions of the ZnDraw client,
            or just the session for the given token.
        timeout : float, optional
            Timeout for the modifier to run in seconds. The Webclient
            will alert the user if the modifier takes longer than this time and
            release the modify button (no further changes are expected, but they
            can happen).
        use_frozen : bool, default=True
            Whether to use the ZnDrawFrozen class to run the modifier.
            The frozen class only allows provides cached data and
            e.g. access to other steps than the current one is not possible.
            If set to false, a full ZnDraw instance will be created for the modifier.
            This can have a performance impact and may lead to timeouts.

        """
        if timeout < 1:
            raise ValueError("Timeout must be at least 1 second")
        if timeout > 300:
            log.critical(
                "Timeout is set to more than 300 seconds. Modifiers might be killed automatically."
            )
        if run_kwargs is None:
            run_kwargs = {}
        if cls.__name__ in self._modifiers:
            raise ValueError(f"Modifier {cls.__name__} already registered")

        self._modifiers[cls.__name__] = {
            "cls": cls,
            "run_kwargs": run_kwargs,
            "public": public,
            "frozen": use_frozen,
            "timeout": timeout,
        }

        _register_modifier(self, self._modifiers[cls.__name__])

    def _run_modifier(self, data: dict):
        self._available = False
        room = data.pop("ZNDRAW_CLIENT_ROOM")
        vis = type(self)(
            url=self.url, token=room, maximum_message_size=self.maximum_message_size
        )
        vis.timeout = self.timeout

        try:
            vis.socket.emit("room:modifier:queue", 0)
            name = data["method"]["discriminator"]

            instance = self._modifiers[name]["cls"](**data["method"])
            instance.run(
                vis,
                timeout=self._modifiers[name]["timeout"],
                **self._modifiers[name]["run_kwargs"],
            )
        except Exception as e:
            log.exception(e)
            vis.log(f"Error: {e}")
        finally:
            vis.socket.emit("room:modifier:queue", -1)

            # wait and then disconnect
            vis.socket.sleep(1)
            vis.socket.disconnect()

            self._available = True
            self.socket.emit("modifier:available", list(self._modifiers))


@dataclasses.dataclass(kw_only=True)
class ZnDrawLocal(ZnDraw):
    """Access database directly.

    This client can / should be used if the database can be accessed directly.
    Data will not be loaded via sockets but modified directly in the database.
    """

    r: Redis

    def __getitem__(self, index: int | list | slice) -> ase.Atoms | list[ase.Atoms]:
        single_item = isinstance(index, int)
        if single_item:
            index = [index]
        if self.r.exists(f"room:{self.token}:frames"):
            data = znsocket.List(self.r, f"room:{self.token}:frames")[index]

        else:
            try:
                data = znsocket.List(self.r, "room:default:frames")[index]
            except IndexError:
                data = []

        structures = [
            znjson.loads(x, cls=znjson.ZnDecoder.from_converters([ASEConverter]))
            for x in data
        ]
        if single_item:
            return structures[0]
        return structures

    def insert(self, index: int, value: ATOMS_LIKE):
        lst = znsocket.List(self.r, f"room:{self.token}:frames")
        if not self.r.exists(f"room:{self.token}:frames"):
            default_lst = znsocket.List(self.r, "room:default:frames")
            # TODO: using a redis copy action would be faster
            lst.extend(default_lst)

        if isinstance(value, ase.Atoms):
            if not hasattr(value, "connectivity") and self.bond_calculator is not None:
                value.connectivity = self.bond_calculator.get_bonds(value)

            value = znjson.dumps(
                value, cls=znjson.ZnEncoder.from_converters([ASEConverter])
            )

        if '"_type": "ase.Atoms"' not in value:
            raise ValueError("Unable to parse provided data object")
        lst.insert(index, value)

    def extend(self, values: list[ATOMS_LIKE]):
        if not isinstance(values, list):
            raise ValueError("Unable to parse provided data object")

        # enable tbar if more than 10 messages are sent
        # approximated by the size of the first frame
        lst = znsocket.List(self.r, f"room:{self.token}:frames")
        show_tbar = (
            len(values)
            * len(
                znjson.dumps(
                    values[0], cls=znjson.ZnEncoder.from_converters([ASEConverter])
                ).encode("utf-8")
            )
        ) > (10 * self.maximum_message_size)
        tbar = tqdm.tqdm(
            values, desc="Sending frames", unit=" frame", disable=not show_tbar
        )

        msg = []

        for val in tbar:
            if isinstance(val, ase.Atoms):
                if not hasattr(val, "connectivity") and self.bond_calculator is not None:
                    val.connectivity = self.bond_calculator.get_bonds(val)

                msg.append(
                    znjson.dumps(
                        val, cls=znjson.ZnEncoder.from_converters([ASEConverter])
                    )
                )
            else:
                msg.append(val)
            if '"_type": "ase.Atoms"' not in msg[-1]:
                raise ValueError("Unable to parse provided data object")
            if len(json.dumps(msg).encode("utf-8")) > self.maximum_message_size:
                lst.extend(msg)
                msg = []
        if len(msg) > 0:  # Only send the message if it's not empty
            lst.extend(msg)

    def __setitem__(
        self,
        index: int | list | slice,
        value: ATOMS_LIKE | list[ATOMS_LIKE],
    ):
        lst = znsocket.List(self.r, f"room:{self.token}:frames")
        if not self.r.exists(f"room:{self.token}:frames"):
            default_lst = znsocket.List(self.r, "room:default:frames")
            # TODO: using a redis copy action would be faster
            lst.extend(default_lst)

        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))
        if isinstance(index, int):
            index = [index]
            value = [value]

        for i, val in zip(index, value):
            if isinstance(val, ase.Atoms):
                if not hasattr(val, "connectivity") and self.bond_calculator is not None:
                    val.connectivity = self.bond_calculator.get_bonds(val)
                val = znjson.dumps(
                    val, cls=znjson.ZnEncoder.from_converters([ASEConverter])
                )
            if '"_type": "ase.Atoms"' not in val:
                raise ValueError("Unable to parse provided data object")
            lst[i] = val



"""zndraw-upload API"""

import eventlet

eventlet.monkey_patch()
import typing as t
import uuid

import typer

from zndraw import ZnDraw

from .tasks import FileIO, get_generator_from_filename

cli = typer.Typer()


def upload(
    filename: str,
    url: str,
    token: t.Optional[str],
    fileio: FileIO,
):
    """Upload a file to ZnDraw."""
    if token is None:
        token = str(uuid.uuid4())
    vis = ZnDraw(url=url, token=token)
    typer.echo(f"Reading {filename} ...")

    generator = get_generator_from_filename(fileio)

    typer.echo(f"Uploading to: {url}/token/{vis.token}")
    vis.extend(list(generator))



import json
import logging
import uuid

import znsocket
from flask import current_app, request, session
from flask_socketio import SocketIO, emit, join_room
from redis import Redis

from zndraw.analyse import Analysis
from zndraw.draw import Geometry
from zndraw.modify import Modifier
from zndraw.scene import Scene
from zndraw.selection import Selection
from zndraw.tasks import (
    run_analysis,
    run_geometry,
    run_modifier,
    run_selection,
    run_upload_file,
)
from zndraw.utils import get_cls_from_json_schema

log = logging.getLogger(__name__)


def init_socketio_events(io: SocketIO):
    @io.on("connect")
    def connect():
        pass

    @io.on("shutdown")
    def shutdown():
        if "AUTH_TOKEN" not in current_app.config or session["authenticated"]:
            log.critical("Shutting down server")
            current_app.extensions["redis"].flushall()

            current_app.extensions["celery"].control.purge()
            current_app.extensions["celery"].control.broadcast("shutdown")

            io.stop()
        else:
            log.critical("Unauthenticated user tried to shut down the server.")

    @io.on("disconnect")
    def disconnect():
        try:
            room = str(session["token"])
        except KeyError:
            log.critical(f"disconnecting {request.sid}")
            return
        r = current_app.extensions["redis"]

        if "name" in session:
            log.critical(f"disconnecting (webclient) {request.sid} from room {room}")
            r.srem(f"room:{room}:webclients", session["name"])
            emit(
                "room:users:refresh", list(r.smembers(f"room:{room}:webclients")), to=room
            )
        else:
            log.critical(f"disconnecting (pyclient) {request.sid}")

            # Get all modifier names associated with the client from Redis
            room_modifier_names = r.hgetall(f"room:{room}:modifiers")
            default_modifier_names = r.hgetall("room:default:modifiers")

            for modifier in room_modifier_names:
                r.srem(f"room:{room}:modifiers:{modifier}", request.sid)
                if r.scard(f"room:{room}:modifiers:{modifier}") == 0:
                    r.hdel(f"room:{room}:modifiers", modifier)
            for modifier in default_modifier_names:
                r.srem(f"room:default:modifiers:{modifier}", request.sid)
                if r.scard(f"room:default:modifiers:{modifier}") == 0:
                    r.hdel("room:default:modifiers", modifier)

            # Emit an event to refresh the modifier schema
            emit("modifier:schema:refresh", broadcast=True)

    @io.on("webclient:connect")
    def webclient_connect():
        try:
            token = session["token"]
        except KeyError:
            token = uuid.uuid4().hex[:8]
            session["token"] = token

        if "AUTH_TOKEN" not in current_app.config:
            session["authenticated"] = True
        else:
            if "authenticated" not in session:
                session["authenticated"] = False

        room = str(session["token"])
        join_room(room)  # rename token to room or room_token

        session["name"] = uuid.uuid4().hex[:8]

        r = current_app.extensions["redis"]
        r.sadd(f"room:{room}:webclients", session["name"])

        # TODO: this is currently not used afaik
        emit("room:users:refresh", list(r.smembers(f"room:{room}:webclients")), to=room)
        # set step, camera, bookmarks, points

        log.critical(f"connecting (webclient) {request.sid} to {room}")

        emit("selection:schema", Selection.updated_schema())
        emit("modifier:schema:refresh")
        emit("scene:schema", Scene.updated_schema())
        emit("geometry:schema", Geometry.updated_schema())
        emit("analysis:schema:refresh")

        # set default arrows config
        if not r.exists(f"room:{room}:config"):
            r.set(
                f"room:{room}:config",
                # TODO: use the default function from config
                json.dumps(
                    {
                        "arrows": {
                            "colormap": [[-0.5, 0.9, 0.5], [0.0, 0.9, 0.5]],
                            "normalize": True,
                            "colorrange": [0, 1],
                            "scale_vector_thickness": False,
                            "opacity": 1.0,
                        },
                    }
                ),
            )

        emit(
            "room:config:set",
            json.loads(r.get(f"room:{room}:config")),
            to=room,
        )

        if "TUTORIAL" in current_app.config:
            emit("tutorial:url", current_app.config["TUTORIAL"])
        if "SIMGEN" in current_app.config:
            emit("showSiMGen", True)

        return {
            "name": session["name"],
            "room": room,
            "authenticated": session["authenticated"],
        }

    @io.on("join")  # rename pyclient:connect
    def join(data: dict):
        """
        Arguments:
            data: {"token": str, "auth_token": str}
        """
        # TODO: prohibit "token" to be "default"

        if "AUTH_TOKEN" not in current_app.config:
            session["authenticated"] = True
        else:
            session["authenticated"] = (
                data["auth_token"] == current_app.config["AUTH_TOKEN"]
            )
        token = data["token"]
        session["token"] = token
        room = str(session["token"])

        join_room(room)
        log.critical(f"connecting (pyclient) {request.sid} to {room}")
        # join_room(f"pyclients_{token}")

    @io.on("room:frames:get")
    def room_frames_get(frames: list[int]) -> dict[int, dict]:
        # print(f"requesting frames: {frames}")
        if len(frames) == 0:
            return {}
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        # TODO: use ZnDrawLocal ?

        if r.exists(f"room:{room}:frames"):
            data = znsocket.List(r, f"room:{room}:frames")[frames]

        else:
            try:
                data = znsocket.List(r, "room:default:frames")[frames]
            except IndexError:
                data = []

        return {idx: json.loads(d) for idx, d in zip(frames, data) if d is not None}

    @io.on("room:frames:set")
    def room_frames_set(data: dict[int, str]):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")

        # add = {}
        # remove = []
        lst = znsocket.List(r, f"room:{room}:frames")

        if not r.exists(f"room:{room}:frames"):
            default_lst = znsocket.List(r, "room:default:frames")
            # TODO: using a redis copy action would be faster
            lst.extend(default_lst)

        data = {int(k): v for k, v in sorted(data.items())}
        for key, value in sorted(data.items()):
            if int(key) == len(lst):
                lst.append(value)
            elif int(key) > len(lst):
                log.critical(f"frame {key} is out of bounds and is being ignored.")
                pass
            else:
                lst[int(key)] = value

        emit("room:frames:refresh", [int(x) for x in data], to=room)
        # This method should be called, because it can move frames from the default
        # room to the current room. Doing so in the background
        # can cause issues with further operations on the frames.
        return "OK"

    @io.on("room:all:frames:refresh")
    def room_all_frames_refresh(indices: list[int]):
        emit("room:frames:refresh", [int(x) for x in indices], broadcast=True)

    @io.on("room:frames:delete")
    def room_frames_delete(frames: list[int]):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        lst = znsocket.List(r, f"room:{room}:frames")
        if not r.exists(f"room:{room}:frames"):
            default_lst = znsocket.List(r, "room:default:frames")
            # TODO: using a redis copy action would be faster
            lst.extend(default_lst)
        del lst[frames]
        # TODO how to update here?
        emit("room:frames:refresh", [int(x) for x in frames], to=room)

        # This method should be called, because it can move frames from the default
        # room to the current room. Doing so in the background
        # can cause issues with further operations on the frames. (see room frames set)
        return "OK"

    @io.on("room:frames:insert")
    def room_frames_insert(data: dict):
        index = data.pop("index")
        value = data.pop("value")
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")

        lst = znsocket.List(r, f"room:{room}:frames")
        if not r.exists(f"room:{room}:frames"):
            default_lst = znsocket.List(r, "room:default:frames")
            # TODO: using a redis copy action would be faster
            lst.extend(default_lst)
        lst.insert(index, value)

        # not sure how to update, insert requires everything to be updated after the insertion
        # can be done custom on the client side to avoid resending everything
        emit("room:frames:refresh", [int(x) for x in data], to=room)
        # This method should be called, because it can move frames from the default
        # room to the current room. Doing so in the background
        # can cause issues with further operations on the frames. (see room frames set)
        return "OK"

    @io.on("room:length:get")
    def room_frames_length_get() -> int:
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]
        room_key = (
            f"room:{room}:frames"
            if r.exists(f"room:{room}:frames")
            else "room:default:frames"
        )
        return len(znsocket.List(r, room_key))

    @io.on("modifier:register")
    def modifier_register(data: dict):
        """Register the modifier."""
        if data["public"] and not session["authenticated"]:
            log.critical("Unauthenticated user tried to register a default modifier.")
            return

        r: Redis = current_app.extensions["redis"]
        room = session.get("token")

        data["ZNDRAW_CLIENT_SID"] = request.sid

        if data.pop("public"):
            r.hset("room:default:modifiers", data["name"], json.dumps(data))
            r.sadd(f"room:default:modifiers:{data['name']}", request.sid)
        else:
            r.hset(f"room:{room}:modifiers", data["name"], json.dumps(data))
            r.sadd(f"room:{room}:modifiers:{data['name']}", request.sid)

        # only if public this is required
        emit("modifier:schema:refresh", broadcast=True)

    @io.on("modifier:schema")
    def modifier_schema():
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")

        modifiers: dict = r.hgetall(f"room:{room}:modifiers")
        modifiers |= r.hgetall("room:default:modifiers")
        # reconstruct them to get the schema
        classes = []
        for modifier in modifiers.values():
            modifier = json.loads(modifier)
            cls = get_cls_from_json_schema(modifier["schema"], modifier["name"])
            classes.append(cls)

        emit(
            "modifier:schema", Modifier.updated_schema(extensions=classes), to=request.sid
        )

    @io.on("draw:schema")
    def draw_schema():
        return Geometry.updated_schema()

    @io.on("scene:schema")
    def scene_schema():
        emit("scene:schema", Scene.updated_schema(), to=request.sid)

    @io.on("selection:schema")
    def selection_schema():
        emit("selection:schema", Selection.updated_schema(), to=request.sid)

    @io.on("analysis:schema")
    def analysis_schema():
        emit("analysis:schema", Analysis.updated_schema(), to=request.sid)

    @io.on("modifier:run")
    def modifier_run(data: dict):
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]

        name = data["method"]["discriminator"]

        public = r.smembers(f"room:default:modifiers:{name}")
        privat = r.smembers(f"room:{room}:modifiers:{name}")

        data["ZNDRAW_CLIENT_ROOM"] = room

        queue_position = 1

        if len(public):
            # The modifier was registered with public=True
            queue_position = r.rpush(f"modifier:queue:{name}", json.dumps(data))
        elif len(privat):
            # The modifier was registered with public=False
            queue_position = r.rpush(f"modifier:queue:{room}:{name}", json.dumps(data))
        else:
            # This would be the queue for default modifiers.
            # but they are queued using celery directly.
            # so no need for redis queue.
            pass

        emit("room:modifier:queue", queue_position, to=room)

        clients: set[str] = public | privat
        if len(clients):
            for sid in clients:
                # kindly ask every client if they are available
                emit("modifier:wakeup", to=sid)
        else:
            run_modifier.delay(room, data)

    @io.on("room:modifier:queue")
    def room_modifier_run(data: int):
        """Forwarding finished message."""
        room = session.get("token")
        emit("room:modifier:queue", data, to=room)

    @io.on("modifier:available")
    def modifier_available(modifier_names: list[str]) -> None:
        """Update state of registered modifier classes."""
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")  # TODO: Why use get, token should always be set

        for name in modifier_names:
            public = r.smembers(f"room:default:modifiers:{name}")
            privat = r.smembers(f"room:{room}:modifiers:{name}")

            if request.sid in public:
                # TODO: !!! there will be an issue if you register a privat modifier "a"
                # and then someone else registers a public modifier "a"
                public_task = r.lpop(f"modifier:queue:{name}")
                if public_task:
                    log.debug(f"running public task {public_task} on {request.sid}")
                    emit("modifier:run", json.loads(public_task), to=request.sid)
                    return

            if request.sid in privat:
                privat_task = r.lpop(f"modifier:queue:{room}:{name}")
                if privat_task:
                    log.debug(f"running private task {privat_task} on {request.sid}")
                    emit("modifier:run", json.loads(privat_task), to=request.sid)
                    return

        log.debug(f"No task available for {modifier_names}")

    @io.on("analysis:run")
    def analysis_run(data: dict):
        room = session.get("token")
        emit("room:analysis:queue", 1, to=room)  # TODO: find the correct queue position
        run_analysis.delay(room, data)

    @io.on("room:analysis:queue")
    def room_analysis_run(data: int):
        """Forwarding finished message."""
        room = session.get("token")
        emit("room:analysis:queue", data, to=room)

    @io.on("room:geometry:queue")
    def room_geometry_run(data: int):
        """Forwarding finished message."""
        room = session.get("token")
        emit("room:geometry:queue", data, to=room)

    @io.on("geometry:run")
    def geometry_run(data: dict):
        room = session.get("token")
        emit("geometry:run:enqueue", to=room)
        run_geometry.delay(room, data)

    @io.on("room:geometry:set")
    def room_geometry_set(data: list):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")

        # # add = {}
        # # remove = []
        lst = znsocket.List(r, f"room:{room}:geometries")
        del lst[:]
        lst.extend(data)
        emit("room:geometry:set", data, to=room, include_self=False)

    @io.on("room:geometry:get")
    def room_geometry_get():
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")

        return list(znsocket.List(r, f"room:{room}:geometries"))

    @io.on("room:config:get")
    def room_config_get():
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        return json.loads(r.get(f"room:{room}:config"))

    @io.on("room:config:set")
    def room_config_set(data: dict):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        config = json.loads(r.get(f"room:{room}:config"))
        config.update(data)
        r.set(f"room:{room}:config", json.dumps(config))
        emit("room:config:set", data, to=room, include_self=False)

    @io.on("analysis:figure:set")
    def analysis_figure_set(data: dict):
        # This is currently using push and the figure is not stored
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]
        r.set(f"room:{room}:analysis:figure", json.dumps(data))
        emit(
            "analysis:figure:set",
            data,
            to=room,
        )

    @io.on("analysis:figure:get")
    def analysis_figure_get() -> dict:
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]
        return json.loads(r.get(f"room:{room}:analysis:figure"))

    @io.on("selection:run")
    def selection_run(data: dict):
        room = session.get("token")
        emit("room:selection:queue", 1, to=room)  # TODO: find the correct queue position
        run_selection.delay(room, data)

    # @io.on("selection:run:finished")
    # def selection_run_finished():
    #     """Forwarding finished message."""
    #     room = session.get("token")
    #     emit("selection:run:finished", to=room)

    # @io.on("selection:run:running")
    # def selection_run_running():
    #     """Forwarding running method."""
    #     room = session.get("token")
    #     emit("selection:run:running", to=room)

    @io.on("room:selection:queue")
    def room_selection_run(data: int):
        """Forwarding finished message."""
        room = session.get("token")
        emit("room:selection:queue", data, to=room)

    @io.on("room:alert")
    def room_alert(msg: str):
        """Forward the alert message to every client in the room"""
        # TODO: identify the source client.
        room = session.get("token")
        emit("room:alert", msg, to=room)

    @io.on("room:log")
    def room_log(msg: str):
        """Forward the alert message to every client in the room"""
        room = session.get("token")
        emit("room:log", msg, to=room)

    @io.on("room:selection:set")
    def room_selection_set(data: dict[str, list[int]]):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        r.hmset(f"room:{room}:selection", {k: json.dumps(v) for k, v in data.items()})
        emit("room:selection:set", data, to=room, include_self=False)

    @io.on("room:selection:get")
    def room_selection_get() -> dict[str, list[int]]:
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        result = r.hgetall(f"room:{room}:selection")
        if "0" in result:
            return {k: json.loads(v) for k, v in result.items()}
        return {"0": []}

    @io.on("room:step:set")
    def room_step_set(step: int):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        r.set(f"room:{room}:step", step)

        emit("room:step:set", step, to=room, include_self=False)

    @io.on("room:step:get")
    def room_step_get() -> int:
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        step = r.get(f"room:{room}:step")
        return int(step) if step else 0

    @io.on("room:points:set")
    def room_points_set(data: dict):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        r.hmset(f"room:{room}:points", {k: json.dumps(v) for k, v in data.items()})

        emit("room:points:set", data, to=room, include_self=False)
        # TODO: add rotation! save position and rotation and scale?

    @io.on("room:points:get")
    def room_points_get() -> dict[str, list[list]]:
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        result: dict[int, list[list]] = r.hgetall(f"room:{room}:points")
        # TODO: type consistency!!
        result = {k: json.loads(v) for k, v in result.items()}
        if "0" not in result:
            return {"0": []}
        return result

    @io.on("room:bookmarks:set")
    def room_bookmarks_set(data: dict):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        if len(data):
            r.hmset(f"room:{room}:bookmarks", data)
        emit("room:bookmarks:set", data, to=room, include_self=False)

    @io.on("room:bookmarks:get")
    def room_bookmarks_get() -> dict:
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        return r.hgetall(f"room:{room}:bookmarks")

    @io.on("room:camera:set")
    def room_camera_set(data: dict):
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        r.set(f"room:{room}:camera", json.dumps(data["content"]))
        if data.get("emit", False):
            emit("room:camera:set", data["content"], to=room, include_self=False)

    @io.on("room:camera:get")
    def room_camera_get() -> dict:
        r: Redis = current_app.extensions["redis"]
        room = session.get("token")
        camera = r.get(f"room:{room}:camera")
        if camera:
            return json.loads(camera)
        return {"position": [-10, -10, -10], "target": [0, 0, 0]}

    @io.on("room:upload:file")
    def room_upload_file(data: dict):
        room = session.get("token")
        run_upload_file.delay(room, data)

    @io.on("room:lock:set")
    def room_lock_set(locked: bool):
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]
        r.set(f"room:{room}:locked", str(locked))
        emit("room:lock:set", locked, to=room)

    @io.on("room:lock:get")
    def room_lock_get() -> bool:
        room = session.get("token")
        r: Redis = current_app.extensions["redis"]
        locked = r.get(f"room:{room}:locked")
        return locked == "True"



import logging
import typing as t
import urllib.request
from io import StringIO

import ase.io
import socketio.exceptions
import tqdm
import znjson
import znsocket
from celery import shared_task
from flask import current_app

from zndraw.base import FileIO
from zndraw.bonds import ASEComputeBonds
from zndraw.exceptions import RoomLockedError
from zndraw.utils import ASEConverter

log = logging.getLogger(__name__)


def _get_default_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    return [ase.Atoms()]


def _get_zntrack_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    node_name, attribute = file_io.name.split(".", 1)
    try:
        import zntrack

        node = zntrack.from_rev(node_name, remote=file_io.remote, rev=file_io.rev)
        images = getattr(node, attribute)
    except ImportError as err:
        raise ImportError(
            "You need to install ZnTrack to use the remote feature."
        ) from err

    return images[file_io.start : file_io.stop : file_io.step]


def _get_znh5md_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    try:
        import znh5md

        io = znh5md.IO(file_io.name)
    except ImportError as err:
        raise ImportError(
            "You need to install ZnH5MD>=0.3 to use the remote feature."
        ) from err
    log.critical(file_io)
    return io[file_io.start : file_io.stop : file_io.step]


def _get_http_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    format = file_io.name.split(".")[-1]
    format = format if format != "xyz" else "extxyz"
    content = urllib.request.urlopen(file_io.name).read().decode("utf-8")
    stream = StringIO(content)

    generator = ase.io.iread(
        stream, format=format, index=slice(file_io.start, file_io.stop, file_io.step)
    )

    return generator


def _get_ase_generator(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    generator = ase.io.iread(
        file_io.name, index=slice(file_io.start, file_io.stop, file_io.step)
    )

    return generator


def get_generator_from_filename(file_io: FileIO) -> t.Iterable[ase.Atoms]:
    if file_io.name is None:
        return _get_default_generator(file_io)
    elif file_io.remote is not None:
        return _get_zntrack_generator(file_io)
    elif file_io.name.endswith((".h5", ".hdf5", ".h5md")):
        return _get_znh5md_generator(file_io)
    elif file_io.name.startswith(("http", "https")):
        return _get_http_generator(file_io)
    else:
        return _get_ase_generator(file_io)


@shared_task
def read_file(fileio: dict) -> None:
    file_io = FileIO(**fileio)
    r = current_app.extensions["redis"]

    io = socketio.Client()
    r.delete("room:default:frames")

    lst = znsocket.List(r, "room:default:frames")
    bonds_calculator = ASEComputeBonds()

    generator = get_generator_from_filename(file_io)

    for idx, atoms in tqdm.tqdm(enumerate(generator)):
        if current_app.config.get("COMPUTE_BONDS", False):
            if not hasattr(atoms, "connectivity"):
                atoms.connectivity = bonds_calculator.get_bonds(atoms)
        lst.append(
            znjson.dumps(atoms, cls=znjson.ZnEncoder.from_converters([ASEConverter]))
        )
        if idx == 0:
            try:
                io.connect(current_app.config["SERVER_URL"], wait_timeout=10)
                io.emit("room:all:frames:refresh", [0])
            except socketio.exceptions.ConnectionError:
                pass

    while True:
        try:
            if not io.connected:
                io.connect(current_app.config["SERVER_URL"], wait_timeout=10)

            # updates len after all frames are loaded
            io.emit("room:all:frames:refresh", [idx])
            break
        except socketio.exceptions.ConnectionError:
            pass

    io.sleep(1)
    io.disconnect()


@shared_task
def run_modifier(room, data: dict) -> None:
    from zndraw.modify import Modifier
    from zndraw.zndraw import ZnDrawLocal

    vis = ZnDrawLocal(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )
    if not current_app.config.get("COMPUTE_BONDS", False):
        vis.bond_calculator = None
    vis.socket.emit("room:modifier:queue", 0)
    try:
        if vis.locked:
            raise RoomLockedError("The room you are trying to modify is locked.")
        modifier = Modifier(**data)
        modifier.run(vis)
    except Exception as e:
        vis.log(str(e))
    finally:
        vis.socket.emit("room:modifier:queue", -1)

    # wait and then disconnect
    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_selection(room, data: dict) -> None:
    from zndraw.selection import Selection
    from zndraw.zndraw import ZnDrawLocal

    vis = ZnDrawLocal(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )
    vis.socket.emit("room:selection:queue", 0)
    try:
        selection = Selection(**data)
        selection.run(vis)
    finally:
        vis.socket.emit("room:selection:queue", -1)

    # wait and then disconnect
    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_analysis(room, data: dict) -> None:
    from zndraw.analyse import Analysis
    from zndraw.zndraw import ZnDrawLocal

    vis = ZnDrawLocal(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )
    vis.socket.emit("room:analysis:queue", 0)
    try:
        analysis = Analysis(**data)
        analysis.run(vis)
    except Exception as e:
        vis.log(str(e))
    finally:
        vis.socket.emit("room:analysis:queue", -1)

    # wait and then disconnect
    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_geometry(room, data: dict) -> None:
    from zndraw.draw import Geometry
    from zndraw.zndraw import ZnDrawLocal

    vis = ZnDrawLocal(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )
    vis.socket.emit("room:geometry:queue", 0)
    try:
        geom = Geometry(**data)
        # TODO: set the position / rotation / scale
        geom.run(vis)
    finally:
        vis.socket.emit("room:geometry:queue", -1)

    # wait and then disconnect
    vis.socket.sleep(1)
    vis.socket.disconnect()


@shared_task
def run_upload_file(room, data: dict):
    from io import StringIO

    import ase.io

    from zndraw.zndraw import ZnDrawLocal

    vis = ZnDrawLocal(
        r=current_app.extensions["redis"],
        url=current_app.config["SERVER_URL"],
        token=room,
    )

    vis.log(f"Uploading {data['filename']}")

    format = data["filename"].split(".")[-1]
    format = format if format != "xyz" else "extxyz"

    if format == "h5":
        raise ValueError("H5MD format not supported for uploading yet")

    stream = StringIO(bytes(data["content"]).decode("utf-8"))

    atoms_list = list(ase.io.iread(stream, format=format))
    if len(atoms_list) == 1 and len(vis.points) != 0:
        scene = vis.atoms
        atoms = atoms_list[0]
        if hasattr(scene, "connectivity"):
            del scene.connectivity
        for point in vis.points:
            atoms.positions -= atoms.get_center_of_mass() - point
            scene += atoms
        vis.append(scene)
    else:
        vis.extend(atoms_list)

    vis.step = len(vis) - 1

    vis.socket.sleep(1)
    vis.socket.disconnect()



"""Base class for all ZnDraw objects."""

import dataclasses


@dataclasses.dataclass
class Material:
    color: str = "#ffffff"
    opacity: float = 1
    wireframe: bool = False


@dataclasses.dataclass
class Object3D:
    position: list[float] | tuple = dataclasses.field(default=(0, 0, 0))
    scale: list[float] | tuple = dataclasses.field(default=(1, 1, 1))
    material: Material = dataclasses.field(default_factory=Material)


@dataclasses.dataclass
class Box(Object3D):
    width: float = 1
    height: float = 1
    depth: float = 1



import eventlet

eventlet.monkey_patch()

import dataclasses
import datetime
import os
import typing as t

import typer

from zndraw.app import create_app
from zndraw.base import FileIO
from zndraw.standalone import run_celery_worker, run_znsocket
from zndraw.tasks import read_file
from zndraw.upload import upload
from zndraw.utils import get_port

cli = typer.Typer()


@dataclasses.dataclass
class EnvOptions:
    FLASK_PORT: str | None = None
    FLASK_STORAGE: str | None = None
    FLASK_AUTH_TOKEN: str | None = None
    FLASK_TUTORIAL: str | None = None
    FLASK_SIMGEN: str | None = None
    FLASK_SERVER_URL: str | None = None
    FLASK_STORAGE_PORT: str | None = None
    FLASK_COMPUTE_BONDS: str | None = None
    FLASK_MAX_HTTP_BUFFER_SIZE: str | None = None

    @classmethod
    def from_env(cls):
        return cls(
            **{
                field.name: os.environ.get(field.name)
                for field in dataclasses.fields(cls)
            }
        )

    def save_to_env(self):
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if value is not None:
                os.environ[field.name] = value


@cli.command()
def main(
    filename: t.Optional[str] = typer.Argument(
        None,
        help="Path to the file which should be visualized in ZnDraw. Can also be the name and attribute of a ZnTrack Node like 'MyNode.atoms' if at least '--remote .' is provided. ",
    ),
    url: t.Optional[str] = typer.Option(
        None,
        help="URL to a running ZnDraw server. Use this server instead of starting a new one.",
    ),
    token: t.Optional[str] = typer.Option(
        None, help="Only valid if 'url' is provided. Room token to upload the file to."
    ),
    port: int = typer.Option(
        None, help="""Port to use for the ZnDraw server. Default port is 1234"""
    ),
    browser: bool = typer.Option(
        True, help="""Whether to open the ZnDraw GUI in the default web browser."""
    ),
    start: int = typer.Option(
        None,
        help="""First frame to be visualized. If set to 0, the first frame will be visualized.""",
    ),
    stop: int = typer.Option(
        None,
        help="""Last frame to be visualized. If set to None, the last frame will be visualized.""",
    ),
    step: int = typer.Option(
        None,
        help="""Stepsize for the frames to be visualized. If set to 1, all frames will be visualized.
        If e.g. set to 2, every second frame will be visualized.""",
    ),
    remote: str = typer.Option(
        None,
        help="URL to a ZnTrack repository to stream data from.",
    ),
    rev: str = typer.Option(
        None,
        help="Revision of the ZnTrack repository to stream data from.",
    ),
    tutorial: str = typer.Option(
        None,
        help="Show the tutorial from the URL inside an IFrame.",
    ),
    auth_token: str = typer.Option(
        None,
        help="Token to authenticate pyclient requests to the ZnDraw server, e.g., for adding defaults to all webclients.",
    ),
    simgen: bool = typer.Option(
        False,
        help="Show the SiMGen demo UI.",
    ),
    storage: str = typer.Option(
        None,
        help="URL to the redis `redis://localhost:6379/0` or znsocket `znsocket://127.0.0.1:6379` server. If None is provided, a local znsocket server will be started.",
    ),
    storage_port: int = typer.Option(
        None, help="Port to use for the storage server. Default port is 6374"
    ),
    standalone: bool = typer.Option(
        True,
        help="Run ZnDraw without additional tools. If disabled, redis and celery must be started manually.",
    ),
    bonds: bool = typer.Option(
        True,
        help="Compute bonds based on covalent distances. This can be slow for large structures.",
    ),
    max_http_buffer_size: int = typer.Option(
        None, help="Maximum size of the HTTP buffer in bytes. Default is 1MB."
    ),
):
    """Start the ZnDraw server.

    Visualize Trajectories, Structures, and more in ZnDraw.
    """
    if token is not None and url is None:
        raise ValueError("You need to provide a URL to use the token feature.")
    if url is not None and port is not None:
        raise ValueError(
            "You cannot provide a URL and a port at the same time. Use something like '--url http://localhost:1234' instead."
        )

    env_config = EnvOptions.from_env()

    if storage_port is not None:
        env_config.FLASK_STORAGE_PORT = str(storage_port)
    elif env_config.FLASK_STORAGE_PORT is None:
        env_config.FLASK_STORAGE_PORT = str(get_port(default=6374))

    if port is not None:
        env_config.FLASK_PORT = str(port)
    elif env_config.FLASK_PORT is None:
        env_config.FLASK_PORT = str(get_port(default=1234))
    if storage is not None:
        env_config.FLASK_STORAGE = storage
    if auth_token is not None:
        env_config.FLASK_AUTH_TOKEN = auth_token
    if tutorial is not None:
        env_config.FLASK_TUTORIAL = tutorial
    if simgen:
        env_config.FLASK_SIMGEN = "TRUE"
    if bonds:
        env_config.FLASK_COMPUTE_BONDS = "TRUE"
    if max_http_buffer_size is not None:
        env_config.FLASK_MAX_HTTP_BUFFER_SIZE = str(int(max_http_buffer_size))

    env_config.FLASK_SERVER_URL = f"http://localhost:{env_config.FLASK_PORT}"

    if standalone and storage is None:
        env_config.FLASK_STORAGE = f"znsocket://localhost:{env_config.FLASK_STORAGE_PORT}"

    env_config.save_to_env()

    if standalone and url is None:
        if env_config.FLASK_STORAGE.startswith("znsocket"):
            # standalone with redis would assume a running instance of redis
            server = run_znsocket(env_config.FLASK_STORAGE_PORT)
        worker = run_celery_worker()

    fileio = FileIO(
        name=filename,
        remote=remote,
        rev=rev,
        start=start,
        stop=stop,
        step=step,
    )

    if url is not None:
        upload(filename, url, token, fileio)
        return

    typer.echo(
        f"{datetime.datetime.now().isoformat()}: Starting zndraw server on port {port}"
    )

    app = create_app()

    read_file.delay(fileio.to_dict())

    if browser:
        import webbrowser

        webbrowser.open(f"http://localhost:{env_config.FLASK_PORT}")

    socketio = app.extensions["socketio"]
    try:
        socketio.run(
            app,
            host="0.0.0.0",
            port=app.config["PORT"],
        )
    finally:
        if standalone:
            server.terminate()
            server.wait()
            print("znsocket server terminated.")
            worker.terminate()
            worker.wait()
            print("celery worker terminated.")



import dataclasses
import typing as t

from zndraw.utils import emit_with_retry

if t.TYPE_CHECKING:
    from zndraw import ZnDraw

HSLColor = t.Tuple[float, float, float]


@dataclasses.dataclass
class ArrowsConfig:
    colormap: list[HSLColor]
    normalize: bool
    colorrange: tuple[float, float]
    scale_vector_thickness: bool
    opacity: float

    _vis = None  # not a dataclass field

    def __setattr__(self, name: str, value) -> None:
        super().__setattr__(name, value)
        if self._vis is not None:
            if name != "vis" and name in [x.name for x in dataclasses.fields(self)]:
                data = {
                    "arrows": dataclasses.asdict(self),
                }
                emit_with_retry(
                    self._vis.socket,
                    "room:config:set",
                    data,
                    retries=self._vis.timeout["emit_retries"],
                )
                self._vis.socket.sleep(0.1)  # maybe use call?

    def set_vis(self, vis: "ZnDraw") -> None:
        self._vis = vis


def _create_arrows_config() -> ArrowsConfig:
    return ArrowsConfig(
        colormap=[[-0.5, 0.9, 0.5], [0.0, 0.9, 0.5]],
        normalize=True,
        colorrange=[0, 1],
        scale_vector_thickness=False,
        opacity=1.0,
    )


@dataclasses.dataclass
class ZnDrawConfig:
    vis: "ZnDraw" = dataclasses.field(repr=False)
    arrows: ArrowsConfig = dataclasses.field(default_factory=_create_arrows_config)

    def __post_init__(self) -> None:
        self.arrows.set_vis(self.vis)



import functools
import importlib.util
import json
import logging
import pathlib
import socket
import sys
import tempfile
import typing as t
import uuid

import ase
import datamodel_code_generator
import numpy as np
import socketio.exceptions
from ase.calculators.singlepoint import SinglePointCalculator
from ase.data import covalent_radii
from ase.data.colors import jmol_colors
from znjson import ConverterBase

from zndraw.type_defs import ASEDict

log = logging.getLogger(__name__)


def rgb2hex(value):
    r, g, b = np.array(value * 255, dtype=int)
    return "#%02x%02x%02x" % (r, g, b)


@functools.lru_cache(maxsize=128)
def get_scaled_radii() -> np.ndarray:
    """Scale down the covalent radii to visualize bonds better."""
    radii = covalent_radii
    # shift the values such that they are in [0.3, 1.3]
    radii = radii - np.min(radii)
    radii = radii / np.max(radii)
    radii = radii + 0.3
    return radii


class ASEConverter(ConverterBase):
    """Encode/Decode datetime objects

    Attributes
    ----------
    level: int
        Priority of this converter over others.
        A higher level will be used first, if there
        are multiple converters available
    representation: str
        An unique identifier for this converter.
    instance:
        Used to select the correct converter.
        This should fulfill isinstance(other, self.instance)
        or __eq__ should be overwritten.
    """

    level = 100
    representation = "ase.Atoms"
    instance = ase.Atoms

    def encode(self, obj: ase.Atoms) -> ASEDict:
        """Convert the datetime object to str / isoformat"""

        numbers = obj.numbers.tolist()
        positions = obj.positions.tolist()
        pbc = obj.pbc.tolist()
        cell = obj.cell.tolist()

        info = {
            k: v
            for k, v in obj.info.items()
            if isinstance(v, (float, int, str, bool, list))
        }
        info |= {k: v.tolist() for k, v in obj.info.items() if isinstance(v, np.ndarray)}
        vectors = info.pop("vectors", [])
        if isinstance(vectors, np.ndarray):
            vectors = vectors.tolist()
        for idx, vector in enumerate(vectors):
            if isinstance(vector, np.ndarray):
                vectors[idx] = vector.tolist()

        if len(vectors) != 0:
            vectors = np.array(vectors)
            if vectors.ndim != 3:
                raise ValueError(
                    f"Vectors must be of shape (n, 2, 3), found '{vectors.shape}'"
                )
            if vectors.shape[1] != 2:
                raise ValueError(
                    f"Vectors must be of shape (n, 2, 3), found '{vectors.shape}'"
                )
            if vectors.shape[2] != 3:
                raise ValueError(
                    f"Vectors must be of shape (n, 2, 3), found '{vectors.shape}'"
                )

            vectors = vectors.tolist()

        if obj.calc is not None:
            calc = {
                k: v
                for k, v in obj.calc.results.items()
                if isinstance(v, (float, int, str, bool, list))
            }
            calc |= {
                k: v.tolist()
                for k, v in obj.calc.results.items()
                if isinstance(v, np.ndarray)
            }
        else:
            calc = {}

        # All additional information should be stored in calc.results
        # and not in calc.arrays, thus we will not convert it here!
        arrays = {}
        if ("colors" not in obj.arrays) or ("" in obj.arrays["colors"]):
            arrays["colors"] = [rgb2hex(jmol_colors[number]) for number in numbers]
        else:
            arrays["colors"] = (
                obj.arrays["colors"].tolist()
                if isinstance(obj.arrays["colors"], np.ndarray)
                else obj.arrays["colors"]
            )

        if ("radii" not in obj.arrays) or (0 in obj.arrays["radii"]):
            # arrays["radii"] = [covalent_radii[number] for number in numbers]
            arrays["radii"] = [get_scaled_radii()[number] for number in numbers]
        else:
            arrays["radii"] = (
                obj.arrays["radii"].tolist()
                if isinstance(obj.arrays["radii"], np.ndarray)
                else obj.arrays["radii"]
            )

        if hasattr(obj, "connectivity") and obj.connectivity is not None:
            connectivity = (
                obj.connectivity.tolist()
                if isinstance(obj.connectivity, np.ndarray)
                else obj.connectivity
            )
        else:
            connectivity = []

        return ASEDict(
            numbers=numbers,
            positions=positions,
            connectivity=connectivity,
            arrays=arrays,
            info=info,
            calc=calc,
            pbc=pbc,
            cell=cell,
            vectors=vectors,
        )

    def decode(self, value: ASEDict) -> ase.Atoms:
        """Create datetime object from str / isoformat"""
        atoms = ase.Atoms(
            numbers=value["numbers"],
            positions=value["positions"],
            info=value["info"],
            pbc=value["pbc"],
            cell=value["cell"],
        )
        if connectivity := value.get("connectivity"):
            # or do we want this to be nx.Graph?
            atoms.connectivity = np.array(connectivity)
        if "colors" in value["arrays"]:
            atoms.arrays["colors"] = np.array(value["arrays"]["colors"])
        if "radii" in value["arrays"]:
            atoms.arrays["radii"] = np.array(value["arrays"]["radii"])
        if calc := value.get("calc"):
            atoms.calc = SinglePointCalculator(atoms)
            atoms.calc.results.update(calc)
        if vectors := value.get("vectors"):
            atoms.info["vectors"] = vectors
        return atoms


def get_port(default: int) -> int:
    """Get an open port."""
    try:
        sock = socket.socket()
        sock.bind(("", default))
        port = default
    except OSError:
        sock = socket.socket()
        sock.bind(("", 0))
        port = sock.getsockname()[1]
    finally:
        sock.close()
    return port


class ZnDrawLoggingHandler(logging.Handler):
    """Logging handler which emits log messages to the ZnDraw server."""

    def __init__(self, zndraw):
        super().__init__()
        self.zndraw = zndraw

    def emit(self, record):
        try:
            msg = self.format(record)
            self.zndraw.log(msg)
        except RecursionError:  # See StreamHandler
            raise
        except Exception:
            print("Something went wrong")
            self.handleError(record)


def get_cls_from_json_schema(schema: dict, name: str, **kwargs):
    """Get a python class from a json schema."""

    # TODO: needs tests
    # TODO: do not write file but use in-memory

    kwargs["strict_nullable"] = True

    with tempfile.TemporaryDirectory() as temporary_directory_name:
        temporary_directory = pathlib.Path(temporary_directory_name)
        output = temporary_directory / "model.py"
        datamodel_code_generator.generate(
            json.dumps(schema),
            input_file_type=datamodel_code_generator.InputFileType.JsonSchema,
            input_filename="example.json",
            output=output,
            # set up the output model types
            output_model_type=datamodel_code_generator.DataModelType.PydanticV2BaseModel,
            **kwargs,
        )

        ref_module = uuid.uuid4().hex

        spec = importlib.util.spec_from_file_location(ref_module, output)
        module = importlib.util.module_from_spec(spec)
        sys.modules[ref_module] = module
        spec.loader.exec_module(module)

        return getattr(module, name)


def emit_with_retry(
    socket: socketio.Client,
    event,
    data=None,
    namespace=None,
    callback=None,
    retries: int = 1,
    delay: float = 0.1,
    increase_delay: float = 0.1,
    reconnect: bool = False,
) -> None:
    """Emit data to a socket with retries."""
    for idx in range(retries):
        try:
            socket.emit(event=event, data=data, namespace=namespace, callback=callback)
            break
        except socketio.exceptions.BadNamespaceError as err:
            log.error(f"Retrying {event} due to {err}")
            if idx == retries - 1:
                raise err
            if reconnect:
                raise ValueError("Reconnect not implemented yet")
        except Exception as err:
            if idx == retries - 1:
                raise err
        socket.sleep(delay)
        delay += increase_delay


def call_with_retry(
    socket: socketio.Client,
    event,
    data=None,
    namespace=None,
    timeout=60,
    retries: int = 1,
    delay: float = 0.1,
    increase_delay: float = 0.1,
    reconnect: bool = False,
) -> t.Any:
    """Call a function with retries."""
    for idx in range(retries):
        try:
            return socket.call(
                event=event, data=data, namespace=namespace, timeout=timeout
            )
        except (
            socketio.exceptions.TimeoutError,
            socketio.exceptions.BadNamespaceError,
        ) as err:
            log.error(f"Retrying {event} due to {err} ({idx} / {retries})")
            if idx == retries - 1:
                raise err
            if reconnect:
                raise ValueError("Reconnect not implemented yet")
        except Exception as err:
            if idx == retries - 1:
                raise err
        socket.sleep(delay)
        delay += increase_delay
    return None


def direction_to_euler(direction, roll=0):
    """
    Convert a direction vector to euler angles.

    You get an increased degree of freedom by setting the roll angle.
    """
    direction = np.array(direction)
    direction = direction / np.linalg.norm(direction)

    x, y, z = direction

    yaw = np.arctan2(y, x)
    pitch = np.arctan2(z, np.sqrt(x**2 + y**2))

    return np.array([yaw, pitch, roll])


def euler_to_direction(angles):
    """
    Convert euler angles to a direction vector.

    Roll gets discarded as this is a reduction in degrees of freedom.
    """
    yaw, pitch, roll = angles

    x = np.cos(yaw) * np.cos(pitch)
    y = np.sin(yaw) * np.cos(pitch)
    z = np.sin(pitch)

    return np.array([x, y, z])
